-- From Falcon Punch 1.0
-- Visit my website for updates at : http://louiscouka.com/

--------------------------------------------------------------------------------
--! Chord
--! Author : Louis Couka
--! Date : 03/03/2016
--------------------------------------------------------------------------------

-------------------------------------------------------------------------------- PIANO CLASS

Piano = {}
Piano.__index = Piano

function Piano:refreshColour(i)
    local colourOff
    local colourOn
    if (isWhite(i-1)) then
        if (self.isNotePlaying[i] > 0) then
            if (self.colourOff[i] == 0) then
                colourOff = "#999999"
            else
                colourOff = "#00bb00"
            end
            colourOn = "#bb0000"
        else
            if (self.colourOff[i] == 0) then
                colourOff = "#ffffff"
            else
                colourOff = "#00ff00"
            end
            colourOn = "#ff0000"
        end
    else
        if (self.isNotePlaying[i] > 0) then
            if (self.colourOff[i] == 0) then
                colourOff = "#000000"
            else
                colourOff = "#009900"
            end
            colourOn = "#990000"
        else
            if (self.colourOff[i] == 0) then
                colourOff = "#444444"
            else
                colourOff = "#00dd00"
            end
            colourOn = "#dd0000"
        end
    end
    self.keyButtons[i].backgroundColourOff = colourOff
    self.keyButtons[i].backgroundColourOn = colourOn
end

function Piano.create(name)
    local piano = {} -- our new object
    setmetatable(piano,Piano) -- make Piano handle lookup
    piano.name = name
    piano.keyButtons = {}
    
    -- Gui Position
    piano.y = 5
    local KEYBOARD_LEFT = 5
    local KEY_WIDTH = 11
    local KEY_BLACK_WIDTH = 6
    local KEY_HEIGHT = 60
    local KEY_BLACK_HEIGHT = 39
    piano.numberKeys = 121
    local FIRST_KEY = 0 -- C-2
    local FIRST_KEY_RANGE = -2
   
    piano.isNotePlaying = {}
    piano.colourOff = {}
    for i = 1,piano.numberKeys do
        piano.isNotePlaying[i] = 0
        piano.colourOff[i] = 0 -- 0 : normal ; 1 : green
    end
  
    -- White Keys
    local note = 0
    local xPos = KEYBOARD_LEFT
    while (note < piano.numberKeys) do
        chroma = note % 12
        i = note + 1
        piano.keyButtons[i] = OnOffButton(name.."_"..tostring(note))
        piano.keyButtons[i].displayName = " "
        piano.keyButtons[i].x = xPos
        piano.keyButtons[i].y = piano.y
        piano.keyButtons[i].width = KEY_WIDTH
        piano.keyButtons[i].height = KEY_HEIGHT
        piano:refreshColour(i)
        if (chroma ~= 4 and chroma ~= 11) then -- E/F and B/C break
            note = note + 2
        else
            note = note + 1
        end
        xPos = xPos + KEY_WIDTH - 1
    end

    -- Black Keys
    local note = 1
    local xPos = KEYBOARD_LEFT + KEY_WIDTH - KEY_BLACK_WIDTH / 2
    while (note < piano.numberKeys) do -- Black
        chroma = note % 12
        i = note + 1
        piano.keyButtons[i] = OnOffButton(name.."_"..tostring(note))
        piano.keyButtons[i].displayName = "1"
        if (chroma == 1 or chroma == 6) then -- Piano asymetries
            piano.keyButtons[i].x = xPos - 1
        elseif (chroma == 10) then
            piano.keyButtons[i].x = xPos + 1
        else
            piano.keyButtons[i].x = xPos
        end
        piano.keyButtons[i].y = piano.y
        piano.keyButtons[i].width = KEY_BLACK_WIDTH
        piano.keyButtons[i].height = KEY_BLACK_HEIGHT
        piano:refreshColour(i)
        if (chroma ~= 3 and chroma ~= 10) then -- E/F and B/C break
            note = note + 2
        else
            note = note + 3
            xPos = xPos + KEY_WIDTH - 1
        end
        xPos = xPos + KEY_WIDTH - 1
    end
    
    piano.stopMouseEventLayer = Label(" ")
    piano.stopMouseEventLayer.x = KEYBOARD_LEFT
    piano.stopMouseEventLayer.y = piano.y
    piano.stopMouseEventLayer.width = 710
    piano.stopMouseEventLayer.height = KEY_HEIGHT
    piano.stopMouseEventLayer.visible = false
    
    return piano
end

function isWhite(note)
    chroma = note % 12
    return chroma ~= 1 and chroma ~= 3 and chroma ~= 6 and chroma ~= 8 and chroma ~= 10
end
    
function Piano:isNoteVisible(note)
    return note < self.numberKeys
end
    
function Piano:setY(y)
    self.y = y
    for i=1, self.numberKeys do
        self.keyButtons[i].y = y
    end
    self.stopMouseEventLayer.y = y
end

function Piano:onNote(note)
    if (note < self.numberKeys) then
        self.isNotePlaying[note+1] = self.isNotePlaying[note+1] + 1
        self:refreshColour(note+1)
    end
end

function Piano:onRelease(note)
    if (note < self.numberKeys and self.isNotePlaying[note+1] > 0) then
        self.isNotePlaying[note+1] = self.isNotePlaying[note+1] - 1
        self:refreshColour(note+1)
    end
end

function Piano:setClickable(isClickable)
    self.stopMouseEventLayer.visible = not isClickable
end

function Piano:setColourOff(i, colourId)
    self.colourOff[i] = colourId
    self:refreshColour(i)
end
piano = Piano.create("Key")
piano:setY(5)

isPlaying = 0
learn = OnOffButton{"Chord_Learn", false, displayName = "Chord Learn", backgroundColourOn = ("#f11a1a"), x = 605, y = 75}
learn.changed = function(self)
    if (self.value) then
        for i = 1,piano.numberKeys do
            if (piano.isNotePlaying[i] > 0) then
                piano.keyButtons[i]:setValue(true)
                piano.isNotePlaying[i] = 0 -- Reset play shadow
                piano:refreshColour(i)
            end
        end        
    end
end

-------------------------------------------------------------------------------- TRANSPOSE STUFF

function transpose(semiton)
    local start, end_, step
    if (semiton < 0) then
        start = 1
        end_ = piano.numberKeys
        step = 1
    elseif (semiton > 0) then
        start = piano.numberKeys
        end_ = 1
        step = -1
    else
        return
    end
    
    for jj=start,end_ + semiton, step do
        piano.keyButtons[jj]:setValue(piano.keyButtons[jj - semiton].value)
    end
    for jj=end_ + semiton,end_ + semiton, step do -- Disable extrems keys
        piano.keyButtons[jj]:setValue(false)
    end
end

function onNote(e)
    isPlaying = isPlaying + 1
    if (learn.value and learn.enabled) then
        playNote(e.note, e.velocity, -1, e.layer, e.channel, e.input, e.vol, e.pan, e.tune, e.slice)
        if (isPlaying == 1) then -- New chord played
            for i = 1,piano.numberKeys do
                piano.keyButtons[i]:setValue(false) -- Reset
            end
        end
        piano.keyButtons[e.note + 1]:setValue(true) -- Add note
    else
        for i, keyButton in ipairs(piano.keyButtons) do
            if keyButton.value then
               playNote(i + e.note - 61, e.velocity, -1, e.layer, e.channel, e.input, e.vol, e.pan, e.tune, e.slice)
            end
        end
    end
    piano:onNote(e.note)
end

function onRelease(e)

    if (isPlaying > 0) then -- failsafe
        isPlaying = isPlaying - 1
    end
    piano:onRelease(e.note)
end

piano.keyButtons[61]:setValue(true)
function loadPreset(preset)
    -- Reset
    for i, keyButton in ipairs(piano.keyButtons) do
        keyButton:setValue(false, false)
    end
    for i = 1, table.getn(preset) do
        piano.keyButtons[preset[i]]:setValue(true, false)
    end
end


local vocingSus4

local vocingMinorMajor7

local vocingMajorMajor7 = {
    
}

local presets =
{
    {
        "--",
        {
            {61}
        },
    },
    
    {
        "Major",
        {
            {49,56,61,65,68,73}, 
            {49,56,61,65,70,75}, 
            {49,56,61,65,68,72,75,80},   
            {49,56,61,63,65,68,70,72,75}, 
            {49,56,61,68,72,77}, 
            {49,56,61,63,65,68}, 
            {56,61,65,68,72,73,75}, 
            {49,56,61,63,65,68,73}, 
            {49,56,61,65,68,70,73}, 
            {49,56,60,63,65,68,72}, 
            {49,56,60,63,65,70}, 
            {49,56,60,63,65,68}, 
            {49,56,61,65,68,72,75}, 
            {49,56,61,65,70,75,80}, 
            {49,53,58,63,68,73}, 
            {53,58,63,68,73,77,80}, 
            {49,53,58,63,68,73,80}, 
            {49,53,58,63,65,68,70}, 
            {49,53,56,58,61,65,68}, 
            {49,53,56,60,63,65,68}, 
            {49,56,61,65,68,72,73,75},
            {49,56,61,65,70,75,77,80},
            {49,56,61,68,73,77,80}, --
            {49,56,65}, 
            {49,56,63,65,68}, 
            {49,56,61,63,65}, --
            {56,61,65,68,73,77}, --
            {53,56,61,65,68,73}, --
            {56,61,65,70,75,77,80}, --  
            {61,63,65,68,73,77,80}, --
            {56,61,65,68,73,77,80,82}, --
            {56,61,63,65,68,70,72,75}, --
            {49,56,61,65,68,72,75,77,80,84}, --
            {49,56,58,61,65,68,72,75,77,80,84,87}, --
        }
    },
    {
        "Minor",
        {
            {49,56,61,64,68,73}, 
            {49,56,61,63,64,68,71}, 
            {49,56,61,68,71,75,78}, 
            {49,56,59,64,68,73}, 
            {49,56,64,68,71,75}, 
            {61,63,64,68,71}, 
            {59,63,64,66,68,71,75,78}, 
            {61,63,64,66,68,71,75}, 
            {49,52,56,61,63,64,68}, 
            {49,52,56,59,64,68,71,75}, 
            {49,56,59,64,68}, 
            {49,56,59,64}, 
            {49,59,64,68,73,78}, 
            {49,56,61,64,68,71}, 
            {49,56,61,64,68,71,75}, 
            {49,56,59,61,64,68}, 
            {49,56,61,64,66,68,71}, 
            {49,56,61,68,71,76}, 
            {49,56,61,64,70,73,75}, 
            {49,56,59,64,68,73,78}, 
            {49,56,64,68,71,73,76},    
            {59,61,64,66,68,71,75,78}, --
            {49,56,59,64,68,73,78,85}, --
            {49,56,61,64,68,73,75}, --
            {49,56,61,64,68,72,75}, --
            {49,52,56,59,63,68,73,80}, --
            {56,59,61,64,66,68,71,73,76}, --
            {49,56,59,63,64,66,71,73,75,78}, --
            {49,52,54,56,59,63,66}, --
            {49,52,56,59,61,64,68}, --
            {49,52,58,63,68,73,78}, --
            {49,58,63,68,73,76,78}, --
            {49,56,59,64,70,75}, --
            {49,51,52,56,59,63}, --
            {49,52,58,63,68,73}, --
            {49,54,59,64,70,75}, --
            {49,56,61,66,71,76}, --
            {52,56,59,63,68,71,76,78,80}, --
            {49,56,61,63,64,68,71,75,78,80,83}, 
        }
    },
    {
        "Dominant",
        {
            --- 7
            {49,56,59,65,68,73}, 
            {49,56,61,65,68,71,73},
            {49,53,59,63,68}, 
            {49,53,59,63,68,73}, 
            {49,59,65,70,75,80}, 
            {59,65,70,75,80,85}, 
            {49,59,65,68,71,75}, 
            {49,59,63,65,68}, 
            
            --- Tensions
            {49,65,69,71,76}, 
            {49,59,65,69,73}, 
            {49,59,62,65,68,71}, 
            {49,56,59,62,67,71}, 
            {49,56,59,62,65,68,73}, 
            {49,56,59,61,62,65,68,71}, 
            {49,56,61,62,67,71,74}, 
            {49,53,59,63,67,70}, 
            {49,53,59,62,65,70}, 
            {49,53,59,62,67,70}, 
            {49,53,59,62,67,71}, 
            {61,63,65,67,69,71,73}, 
            {49,59,62,65,67,71}, 
            {49,59,62,65,71}, 
            {49,59,64,65,68}, 
            {49,53,59,64,68,73}, 
            {49,56,62,65,67,68,71,73}, 
            {49,56,59,61,62,65,68,71,74}, 
            {49,56,59,62,65,69}, 
            {49,56,61,65,69,71,76}, 
            {49,53,59,65,69,73}, 
            {49,56,61,65,69,71,74,76}, 
            {53,59,64,68,73,80}, 
            {49,56,59,65,67,69,71,73}, 
        }
    },
    {
        "Sus4",
        {
            {49,56,59,63,66,70,73}, 
            {49,56,59,61,63,66,68,73}, 
            {49,56,59,63,66,68,71}, 
            {49,56,59,63,66,70}, 
            {49,56,59,61,63,66,70,75}, 
            {49,56,59,61,63,66,68,71,73}, 
            {54,56,59,61,63,66,70,73}, 
            {49,54,59,63,66,70}, 
            {49,54,56,58,59,63}, 
            {49,56,63,66,68,71}, 
            {49,56,59,66,68,71,73,75}, 
            {49,56,61,68,71,75,78}, 
            {49,56,59,63,66,68}, 
            {49,56,59,63,66,71,73}, 
            {49,56,59,63,66,68,71,73}, 
            {56,59,61,63,66,68,71,73,75,78}, 
            {49,59,63,66,70,73}, 
            {49,59,63,66,70,73,75,78,80}, 
            {61,63,66,68,71,73,75,78}, 
            {54,56,59,61,63,66,68,70,73}, 
            {51,54,56,59,61,66,68,71,73}, 
            {49,54,59,63,66,71,75}, 
            {49,54,56,58,59,63,66}, 
            {61,66,68,71,75,78}, 
            {61,63,66,68,71,75}, 
            {61,66,68,70,71,75}, 
            {61,66,68,70,71,75,78}, 
            {49,54,59,63,66,71}, 
            {49,56,59,61,63,66,68,71,73,75,78,80}, 
            {49,54,59,63,68,73}, 
            {49,54,59,61,63,68,73,78}, 
            {49,54,56,59,61,64,66,68,71,73,78}, 
            {49,56,59,63,68,71,75}, 
        }
    }
}

---- UNCOMMENT THIS TO SEE IF THERE ARE DOUBLONS
-- function compareTable(t1, t2)
--     if table.getn(t1) == table.getn(t2) then
--         for i, note in ipairs(preset) do
--             if t1[i] ~= t2[i] then return false end
--         end
--         return true
--     else
--         return false
--     end
-- end
-- 
-- for i, preset_ in ipairs(presets) do
--     print(preset_[1])
--     preset = preset_[2]
--     for j, note in ipairs(preset) do
--         for k, note in ipairs(preset) do
--             if j ~= k and compareTable(preset[j], preset[k]) then print(tostring(j).."="..tostring(k)) end
--         end
--     end
-- end

presetNames = {}
for i = 1, table.getn(presets) do
    presetNames[i] = presets[i][1]
end

coloursMenu = Menu{"Colours", presetNames, showLabel = false, height = 20, x = 5, y = 75, width = 100}
vocing = NumBox{"Voicing", 1, 1, 99, true, x = coloursMenu.x + coloursMenu.width + 9, y = 75, width = coloursMenu.width}
function loadSelectedPreset()
    loadPreset(presets[coloursMenu.value][2][vocing.value])
end
vocing.changed = function(self)
    numVoicings = table.getn(presets[coloursMenu.value][2])
    if self.value > numVoicings then self:setValue(numVoicings, false) end
    self.displayText = tostring(self.value).."/"..tostring(numVoicings)
    loadSelectedPreset()
end
coloursMenu.changed = function(self)
    vocing:setValue(1, false)
    vocing.changed(vocing)
end
vocing.changed(vocing)

voicingUp = Button{"Voicing_Up", displayName = "+", width = 15, height = 12, x = vocing.x + vocing.width + 3, y = vocing.y - 1}
voicingUp.changed = function(self) vocing:setValue(vocing.value+1) end
voicingDown = Button{"Voicing_Down", displayName = "-", width = 15, height = 12, x = vocing.x + vocing.width + 3, y = vocing.y + 10}
voicingDown.changed = function(self) vocing:setValue(vocing.value-1) end

